/**
 *
 * @author  kjdon@cs.waikato.ac.nz
 * @author  davidb@cs.waikato.ac.nz
 * @author  michael@dlconsulting.com
 * @author  john@dlconsulting.com
 * @version 
 */

package org.nzdl.gsdl.LuceneWrap;


import java.io.*;
import java.util.*;
import java.util.regex.*;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.TermDocs;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.BooleanQuery.TooManyClauses;
import org.apache.lucene.search.Filter;
import org.apache.lucene.search.Hit;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.RangeFilter;
import org.apache.lucene.search.Searcher;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.Sort;
import org.apache.lucene.search.TopFieldDocs;


public class GS2LuceneQuery
{
    static private String TEXTFIELD = "TX";

    // Use the standard set of English stop words by default
    static private String[] stop_words = StandardAnalyzer.STOP_WORDS;

    // Command-line options
    static private String fuzziness = null;
    static private Filter filter = null;
    static private Sort sorter = new Sort();
    static private String default_conjuction_operator = "OR";
    static private int start_results = 1;
    static private int end_results = Integer.MAX_VALUE;


    static public void main (String args[])
    {
	if (args.length == 0) {
	    System.out.println("Usage: GS2LuceneQuery <index directory> [-fuzziness value] [-filter filter_string] [-sort sort_field] [-dco AND|OR] [-startresults number -endresults number]");
	    return;
	}

        try {
	    Searcher searcher = new IndexSearcher(args[0]);
	    IndexReader reader = ((IndexSearcher) searcher).getIndexReader();

	    // Create one query parser with the standard set of stop words, and one with none
	    QueryParser query_parser = new QueryParser(TEXTFIELD, new StandardAnalyzer(stop_words));
	    QueryParser query_parser_no_stop_words = new QueryParser(TEXTFIELD, new StandardAnalyzer(new String[] { }));

            for (int i = 1; i < args.length; i++) {
		if (args[i].equals("-sort")) {
		    i++;
		    sorter = new Sort(args[i]);
		}
		else if (args[i].equals("-filter")) {
		    i++;
		    filter = parseFilterString(args[i]);
		}
		else if (args[i].equals("-dco")) {
		    i++;
		    default_conjuction_operator = args[i];
		}
		else if (args[i].equals("-fuzziness")) {
		    i++;
		    fuzziness = args[i];
		}
		else if (args[i].equals("-startresults")) {
		    i++;
		    if (args[i].matches("\\d+")) {
			start_results = Integer.parseInt(args[i]);
		    }
		}
		else if (args[i].equals("-endresults")) {
		    i++;
		    if (args[i].matches("\\d+")) {
			end_results = Integer.parseInt(args[i]);
		    }
		}
	    }
	    
	    // Lucene does "OR" queries by default; do an "AND" query if specified
	    if (default_conjuction_operator.equals("AND")) {
		query_parser.setDefaultOperator(query_parser.AND_OPERATOR);
		query_parser_no_stop_words.setDefaultOperator(query_parser.AND_OPERATOR);
	    }

	    BufferedReader in = new BufferedReader(new InputStreamReader(System.in, "UTF-8"));
	    while (true) {
		// Read the query from STDIN
		String query_string = in.readLine();
		if (query_string == null || query_string.length() == -1) {
		    break;
		}
		System.out.println("<ResultSet>");
		System.out.println("  <QueryString>" + xmlSafe(query_string) + "</QueryString>");
		if (filter != null) {
		    System.out.println("  <FilterString>" + filter.toString() + "</FilterString>");
		}

		try {
		    Query query_including_stop_words = parseQuery(reader, query_parser_no_stop_words, query_string, fuzziness);
		    query_including_stop_words = query_including_stop_words.rewrite(reader);

		    Query query = parseQuery(reader, query_parser, query_string, fuzziness);
		    query = query.rewrite(reader);

		    // Return the list of expanded query terms and their frequencies
		    HashSet terms = new HashSet();
		    query.extractTerms(terms);
		    Iterator term_iterator = terms.iterator();
		    System.out.println("  <QueryTermsInfo num=\"" + terms.size() + "\"/>");
		    while (term_iterator.hasNext()) {
			Term term = (Term) term_iterator.next();

			// Get the term frequency over all the documents
			TermDocs term_docs = reader.termDocs(term);
			int term_freq = term_docs.freq();
			while (term_docs.next()) {
			    term_freq += term_docs.freq();
			}

			// If you wanted to limit this to just text terms add
			// something like this:
			// if (term.field().equals(TEXTFIELD))
			System.out.println("  <Term value=\"" + term.text() + "\" field=\"" + term.field() + "\" freq=\"" + term_freq + "\" />");
		    }
		    
		    // Return the list of stop words removed from the query
		    HashSet terms_including_stop_words = new HashSet();
		    query_including_stop_words.extractTerms(terms_including_stop_words);
		    Iterator terms_including_stop_words_iter = terms_including_stop_words.iterator();
		    while (terms_including_stop_words_iter.hasNext()) {
			Term term = (Term) terms_including_stop_words_iter.next();
			if (!terms.contains(term)) {
			    System.out.println("  <StopWord value=\"" + term.text() + "\"/>");
			}
		    }

		    // Simple case for getting all the matching documents
		    if (end_results == Integer.MAX_VALUE) {
			// Perform the query (filter and sorter may be null)
			Hits hits = searcher.search(query, filter, sorter);
			System.out.println("  <MatchingDocsInfo num=\"" + hits.length() + "\"/>");

			// Output the matching documents
			System.out.println("  <StartResults num=\"" + start_results + "\" />");
			System.out.println("  <EndsResults num=\"" + hits.length() + "\" />");
			for (int i = start_results; i <= hits.length(); i++) {
			    Document doc = hits.doc(i - 1);
 			    System.out.println("  <Match id=\"" + doc.get("nodeID") + "\" />");
			}
		    }

		    // Slightly more complicated case for returning a subset of the matching documents
		    else {
			// Perform the query (filter may be null)
			TopFieldDocs hits = searcher.search(query, filter, end_results, sorter);
			System.out.println("  <MatchingDocsInfo num=\"" + hits.totalHits + "\"/>");

			// Output the matching documents
			System.out.println("  <StartResults num=\"" + start_results + "\" />");
			System.out.println("  <EndsResults num=\"" + end_results + "\" />");
			for (int i = start_results; (i <= hits.scoreDocs.length && i <= end_results); i++) {
			    Document doc = reader.document(hits.scoreDocs[i - 1].doc);
 			    System.out.println("  <Match id=\"" + doc.get("nodeID") + "\" />");
			}
		    }
		}
		catch (ParseException parse_exception) {
		    System.out.println("  <Error type=\"PARSE_EXCEPTION\"/>");
		}
		catch (TooManyClauses too_many_clauses_exception) {
		    System.out.println("  <Error type=\"TOO_MANY_CLAUSES\"/>");
		}

		System.out.println("</ResultSet>");
	    }
	    
	    searcher.close();
	}
	catch (IOException exception) {
	    exception.printStackTrace();
	}
    }

    private static String xmlSafe(String text) {
        return text.replaceAll("\\&", "\\&amp;");
    }

    private static Query parseQuery(IndexReader reader, QueryParser query_parser, String query_string, String fuzziness)
	throws java.io.IOException, org.apache.lucene.queryParser.ParseException
    {
	// Split query string into the search terms and the filter terms
	// * The first +(...) term contains the search terms so count
	//   up '(' and stop when we finish matching ')'
	int offset = 0;
	int paren_count = 0;
	boolean seen_paren = false;
	while (offset < query_string.length() && (!seen_paren || paren_count > 0)) {
	    if (query_string.charAt(offset) == '(') {
		paren_count++;
		seen_paren = true;
	    }
	    if (query_string.charAt(offset) == ')') {
		paren_count--;
	    }
	    offset++;
	}
	String query_prefix = query_string.substring(0, offset);
	String query_suffix = query_string.substring(offset);

	///ystem.err.println("Prefix: " + query_prefix);
	///ystem.err.println("Suffix: " + query_suffix);

	Query query = query_parser.parse(query_prefix);
	query = query.rewrite(reader);

	// If this is a fuzzy search, then we need to add the fuzzy
	// flag to each of the query terms
	if (fuzziness != null && query.toString().length() > 0) {
	    // Revert the query to a string
	    System.err.println("Rewritten query: " + query.toString());
	    // Search through the string for TX:<term> query terms
	    // and append the ~ operator. Not that this search will
	    // not change phrase searches (TX:"<term> <term>") as
	    // fuzzy searching is not possible for these entries.
	    // Yahoo! Time for a state machine!
	    StringBuffer mutable_query_string = new StringBuffer(query.toString());
	    int o = 0; // Offset
	    // 0 = BASE, 1 = SEEN_T, 2 = SEEN_TX, 3 = SEEN_TX:
	    int s = 0; // State
	    while (o < mutable_query_string.length()) {
		char c = mutable_query_string.charAt(o);
		if (s == 0 && c == TEXTFIELD.charAt(0)) {
		    ///ystem.err.println("Found T!");
		    s = 1;
		}
		else if (s == 1) {
		    if (c == TEXTFIELD.charAt(1)) {
			///ystem.err.println("Found X!");
			s = 2;
		    }
		    else {
			s = 0; // Reset
		    }
		}
		else if (s == 2) {
		    if (c == ':') {
			///ystem.err.println("Found TX:!");
			s = 3;
		    }
		    else {
			s = 0; // Reset
		    }
		}
		else if (s == 3) {
		    // Don't process phrases
		    if (c == '"') {
			///ystem.err.println("Stupid phrase...");
			s = 0; // Reset
		    }
		    // Found the end of the term... add the
		    // fuzzy search indicator
		    // Nor outside the scope of parentheses
		    else if (Character.isWhitespace(c) || c == ')') {
			///ystem.err.println("Yahoo! Found fuzzy term.");
			mutable_query_string.insert(o, '~' + fuzziness);
			o++;
			s = 0; // Reset
		    }
		}
		o++;
	    }
	    // If we were in the state of looking for the end of a
	    // term - then we just found it!
	    if (s == 3) {
		mutable_query_string.append('~' + fuzziness);
	    }
	    // Reparse the query
	    ///ystem.err.println("Fuzzy query: " + mutable_query_string.toString() + query_suffix);
	    query = query_parser.parse(mutable_query_string.toString() + query_suffix);
	}
	else {
	    query = query_parser.parse(query_prefix + query_suffix);
	}
	
	return query;
    }


  /**
   *  @todo Michael to comment
   */
  private static Filter parseFilterString(String filter_string)
  {
    Filter result = null;
    Pattern pattern = Pattern.compile("\\s*\\+(\\w+)\\:([\\{\\[])(\\d+)\\s+TO\\s+(\\d+)([\\}\\]])\\s*");
    Matcher matcher = pattern.matcher(filter_string);
    if (matcher.matches()) {
        String field_name = matcher.group(1);
        boolean include_lower = matcher.group(2).equals("[");
        String lower_term = matcher.group(3);
        String upper_term = matcher.group(4);
        boolean include_upper = matcher.group(5).equals("]");
        result = new RangeFilter(field_name, lower_term, upper_term, include_lower, include_upper);
    }
    else {
        System.err.println("Error: Could not understand filter string \"" + filter_string + "\"");
    }
    return result;
  }
  /** parseFilterString() **/
}
